"""
GPIO task for DistKV
"""

import anyio
import asyncgpio

from distkv_ext.gpio.model import GPIOchip

import logging

logger = logging.getLogger(__name__)


async def task(chip: GPIOchip, evt=None):
    with asyncgpio.open_chip(label=chip.name) as srv:
        try:
            async with anyio.create_task_group() as tg:
                chip.task_group = tg
                await chip.set_chip(srv)
                if evt is not None:
                    await evt.set()

                while True:
                    await anyio.sleep(99999)
        finally:
            chip.task_group = None



from distkv.util import attrdict, Path

CFG = attrdict(
    prefix=Path(".distkv", "gpio"),
    interval=1,  # input/count: Pulse reporting frequency
    count=True,  # input/count, input/button: Pulse direction
    # up down both=True False None
    low=False,  # if True, the default is active-low
    flow=False,  # if True, send intermediate results
    skip=True,  # ignore signals < t_bounce
    t_idle=1.5,  # input/button: max pulse width
    t_clear=30,  # button: set to None after this
    t_bounce=0.05,  # input/button: min pulse width
)



import sys
import anyio
import trio
import trio.testing

from distmqtt.test import test_client
from asyncgpio.test import GpioWatcher, Pin

from distkv_ext.gpio.task import task as GPIOtask
from distkv_ext.gpio.model import GPIOroot
from distkv_ext.gpio.config import CFG
from distkv.util import Path

import logging

logger = logging.getLogger(__name__)

PinIH = Pin(False, True)
PinIL = Pin(False, False)
PinOH = Pin(True, True)
PinOL = Pin(True, False)


async def fwd_q(ts, queues):
    while True:
        msg = await ts.__anext__()
        await queues[msg.path[-1]].put(msg.value)


tests = {}
only = ()
# only = ("test_four","test_four_only",)
# only = ("test_three_bounce_only",)


class _test_m(type):
    def __new__(cls, name, bases, classdict):
        result = type.__new__(cls, name, bases, classdict)
        if result.pin is None:
            return result
        if only and name not in only:
            return result
        try:
            tl = tests[result.pin]
        except KeyError:
            tl = tests[result.pin] = []
        tl.append(result)
        return result


class _test(metaclass=_test_m):
    client = None
    queue = None
    pin = None
    prep = None
    typ = None  # overridden

    def add_prep(self, pr):
        pass

    async def prepare(self, c, host, label):
        pr = {"type": self.typ}
        pr.update(self.prep)
        self.add_prep(pr)
        for s in ("src", "dest", "state"):
            if s in pr:
                setattr(self, s, pr[s])
        return await c.set(CFG.prefix + (host, label, self.pin), value=pr, nchain=1)

    async def task(self, c, q, p):
        self.client = c
        self.queue = q
        self.pin = p
        await self.run()  # pylint: disable=no-member

    async def flushMsgs(self, *, timeout=1):
        """
        Flush the message queue, ensure that we're not getting flooded
        """
        async with anyio.move_on_after(timeout):
            while True:
                await self.queue.get()
        try:
            async with anyio.fail_after(timeout / 10):
                msg = await self.queue.get()
        except TimeoutError:
            pass
        else:
            assert False, msg

    async def assertMsg(self, *data, timeout=1, pick=None):
        """
        Check that the incoming messages match what we expect.
        """
        want_pins = []
        want_msgs = []
        in_pins = []
        in_msgs = []
        # Sort pin changes and DistKV messages into different bins
        # because they may interleave randomly
        for d in data:
            if isinstance(d, Pin):
                want_pins.append(d)
            else:
                want_msgs.append(d)
        try:
            while True:
                async with anyio.fail_after(timeout):
                    msg = await self.queue.get()
                    if isinstance(msg, Pin):
                        in_pins.append(msg)
                        if len(in_pins) > len(want_pins):
                            # protect against getting flooded
                            break
                    else:
                        if pick:
                            try:
                                msg = msg[pick]
                            except KeyError:
                                raise KeyError((msg, set)) from None
                        in_msgs.append(msg)
                        if len(in_msgs) > len(want_msgs):
                            # protect against getting flooded
                            break
        except TimeoutError:
            pass
        assert in_msgs == want_msgs, (in_msgs, want_msgs)
        assert in_pins == want_pins, (in_pins, want_pins)


class _test_in(_test):
    typ = "input"
    dest = None  # prep

    def add_prep(self, pr):
        super().add_prep(pr)
        pr["dest"] = ("test", "state", self.pin)


class _test_out(_test):
    typ = "output"
    src = None  # prep
    state = None  # prep

    def add_prep(self, pr):
        super().add_prep(pr)
        pr["src"] = ("test", "gpio", self.pin)
        pr["state"] = ("test", "state", self.pin)

    async def set_src(self, value):
        await self.client.set(self.src, value=value)


class test_one(_test_in):
    pin = 1
    prep = dict(mode="read")

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs()

        await self.assertMsg()
        self.pin.set(False)
        await self.assertMsg()
        self.pin.set(True)
        await self.assertMsg(PinIH, True)
        self.pin.set(False)
        await self.assertMsg(PinIL, False)
        self.pin.set(False)
        await self.assertMsg()


class test_one_uniq(_test_in):
    pin = 1
    prep = dict(mode="read", change=True)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs()

        await self.assertMsg()
        self.pin.set(False)
        await self.assertMsg()
        self.pin.set(True)
        await self.assertMsg(PinIH, True)
        self.pin.set(True)
        await self.assertMsg()
        self.pin.set(False)
        await self.assertMsg(PinIL)
        self.pin.set(False)
        await self.assertMsg()


class test_two(_test_in):
    pin = 2
    prep = dict(mode="count", interval=3, count=None)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs()

        await self.assertMsg()
        await self.client.set(self.dest, value=0)
        await self.assertMsg(0)
        self.pin.set(True)
        await self.assertMsg(PinIH, 1, timeout=0.3)
        await self.assertMsg(timeout=0.3)
        self.pin.set(False)
        await self.assertMsg(PinIL, timeout=0.3)
        self.pin.set(True)
        await self.assertMsg(PinIH, timeout=0.3)
        self.pin.set(False)
        await self.assertMsg(PinIL, timeout=0.3)
        await self.assertMsg(4, timeout=2)
        await self.assertMsg(timeout=5)
        self.pin.set(True)
        await self.assertMsg(PinIH, 5, timeout=0.3)
        await self.assertMsg(timeout=5)


class test_two_up(_test_in):
    pin = 2
    prep = dict(mode="count", interval=3, count=True)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs()

        await self.assertMsg()
        await self.client.set(self.dest, value=0)
        await self.assertMsg(0)
        self.pin.set(True)
        await self.assertMsg(PinIH, 1, timeout=0.3)
        await self.assertMsg(timeout=0.3)
        self.pin.set(False)
        await self.assertMsg(PinIL, timeout=0.3)
        self.pin.set(True)
        await self.assertMsg(PinIH, timeout=0.3)
        self.pin.set(False)
        await self.assertMsg(PinIL, timeout=0.3)
        await self.assertMsg(2, timeout=5)
        await self.assertMsg(timeout=5)


class test_three(_test_in):
    pin = 3
    prep = dict(mode="button", t_idle=1, t_bounce=0.1)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs(timeout=1.2)

        await self.assertMsg()
        await self.client.set(self.dest, value=None)
        await self.assertMsg(None)

        self.pin.set(True)
        await self.assertMsg(PinIH, timeout=0.4)
        self.pin.set(False)
        await self.assertMsg(PinIL, timeout=0.9)
        await self.assertMsg((4, 0), timeout=0.5, pick="seq")
        await self.assertMsg()


class test_three_bounce(_test_in):
    pin = 3
    prep = dict(mode="button", t_idle=1, t_bounce=0.2)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs(timeout=1.2)

        await self.assertMsg()
        await self.client.set(self.dest, value=None)
        await self.assertMsg(None)

        self.pin.set(True)
        for _ in range(3):
            await self.assertMsg(PinIH, timeout=0.15)
            self.pin.set(False)
            await self.assertMsg(PinIL, timeout=0.15)
            self.pin.set(True)
        await self.assertMsg(PinIH, timeout=0.35)
        self.pin.set(False)
        for _ in range(3):
            await self.assertMsg(PinIL, timeout=0.15)
            self.pin.set(True)
            await self.assertMsg(PinIH, timeout=0.15)
            self.pin.set(False)
        await self.assertMsg(PinIL, timeout=0.9)
        await self.assertMsg((6, 0), timeout=0.3, pick="seq")


class test_three_bounce_only(_test_in):
    pin = 3
    prep = dict(mode="button", t_idle=1, t_bounce=0.2)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs(timeout=1.2)

        await self.assertMsg()
        await self.client.set(self.dest, value=None)
        await self.assertMsg(None)

        for _ in range(6):
            self.pin.set(True)
            await self.assertMsg(PinIH, timeout=0.15)
            self.pin.set(False)
            await self.assertMsg(PinIL, timeout=0.15)
        await self.assertMsg(timeout=2)


class test_three_bounce_skip(_test_in):
    pin = 3
    prep = dict(mode="button", t_idle=1, t_bounce=0.2, skip=False)

    async def run(self):
        self.pin.set(False)
        await self.flushMsgs(timeout=1.2)

        await self.assertMsg()
        await self.client.set(self.dest, value=None)
        await self.assertMsg(None)

        for _ in range(6):
            self.pin.set(True)
            await self.assertMsg(PinIH, timeout=0.15)
            self.pin.set(False)
            await self.assertMsg(PinIL, timeout=0.15)
        await self.assertMsg((8, 0), timeout=2, pick="seq")


class test_four(_test_out):
    pin = 4
    prep = dict(mode="write")

    async def run(self):
        await self.set_src(False)
        await self.flushMsgs()

        await self.set_src(True)
        await self.assertMsg(PinOH, True)
        await self.set_src(True)
        await self.assertMsg(True)
        await self.set_src(False)
        await self.assertMsg(PinOL, False)
        await self.set_src(False)
        await self.assertMsg(False)


class test_four_only(_test_out):
    pin = 4
    prep = dict(mode="write", change=True)

    async def run(self):
        await self.set_src(False)
        await self.flushMsgs()

        await self.set_src(True)
        await self.assertMsg(PinOH, True)

        await self.client.set(self.state, value=None)
        await self.assertMsg(None)

        await self.set_src(False)
        await self.assertMsg(PinOL)
        await self.set_src(False)
        await self.assertMsg()
        await self.set_src(True)
        await self.assertMsg(PinOH, True)
        await self.set_src(False)
        await self.assertMsg(PinOL)


class test_five(_test_out):
    pin = 5
    prep = dict(mode="oneshot", t_on=1)

    async def run(self):
        await self.set_src(False)
        await self.flushMsgs()

        await self.set_src(True)
        await self.assertMsg(PinOH, True, timeout=0.4)
        await self.assertMsg(timeout=0.3)
        await self.assertMsg(PinOL, False)
        await self.set_src(False)
        await self.assertMsg()

        await self.set_src(True)
        await self.assertMsg(PinOH, True, timeout=0.3)
        await self.set_src(False)
        await self.assertMsg(PinOL, False, timeout=0.3)
        await self.assertMsg()


class test_six(_test_out):
    pin = 6
    prep = dict(mode="pulse", t_on=1, t_off=3)

    async def run(self):
        await self.set_src(False)
        await self.flushMsgs()

        await self.set_src(True)
        await self.assertMsg(PinOH, 0.25, timeout=0.4)
        await self.assertMsg(timeout=0.4)
        await self.assertMsg(PinOL, timeout=0.4)
        await self.set_src(False)
        await self.assertMsg(False, timeout=0.2)
        await self.assertMsg()

        await self.set_src(True)
        await self.assertMsg(PinOH, 0.25, timeout=0.3)
        await self.set_src(False)
        await self.assertMsg(PinOL, False, timeout=0.3)
        await self.assertMsg()

        await self.set_src(True)
        await self.assertMsg(PinOH, 0.25, timeout=0.4)
        for _ in range(3):
            await self.assertMsg(timeout=0.3)
            await self.assertMsg(PinOL, timeout=0.5)
            await self.assertMsg(timeout=2.3)
            await self.assertMsg(PinOH, timeout=0.5)
        await self.set_src(False)
        await self.assertMsg(PinOL, False, timeout=0.3)
        await self.assertMsg()


async def main(label="gpio-mockup-A", host="HosT"):
    logging.basicConfig(level=logging.DEBUG, format="%(relativeCreated)d %(name)s %(message)s")

    async with test_client() as c, GpioWatcher(interval=0.05).run() as w, c.watch(
        Path("test","state")
    ) as ts:
        ts = ts.__aiter__()  # currently a NOP but you never know
        server = await GPIOroot.as_handler(c)
        await server.wait_loaded()

        controller = server.follow(Path(host, label), create=None)

        async with anyio.create_task_group() as tg:
            evt = anyio.create_event()
            await tg.spawn(GPIOtask, controller, evt)
            await evt.wait()

            try:
                while True:
                    async with anyio.fail_after(1):
                        msg = await ts.__anext__()
                        print("init", msg)
            except TimeoutError:
                pass
            await anyio.sleep(1)

            async def watcher(q, p):
                # Monitor changes of the pin and forward them to the queue
                async with p.watch() as pq:
                    async for m in pq:
                        await q.put(Pin(*m))

            async def runner(tl, c, q, p):
                # run all tests in TL
                async with anyio.create_task_group() as tj:
                    await tj.spawn(watcher, q, p)
                    for t in tl:
                        if isinstance(t, type):
                            t = t()
                        res = await t.prepare(c, host, label)
                        await server.wait_chain(res.chain)
                        await t.task(c, q, p)
                    await tj.cancel_scope.cancel()

            async with anyio.create_task_group() as tt:
                queues = {}
                for nr in tests:
                    queues[nr] = anyio.create_queue(10)
                await tg.spawn(fwd_q, ts, queues)
                for nr, tl in tests.items():
                    await tt.spawn(runner, tl, c, queues[nr], w.pin(label, nr))
            # we come here when all tests have finished
            await tg.cancel_scope.cancel()

            pass  # wait for TG end

        found = 0
        for err in server.err.all_errors():
            if err.resolved:
                continue
            found += 1
            logger.error(
                "Err %s", " ".join(str(x) for x in err.path),
            )
            for e in err:
                logger.error("%s: %r", e.comment, e.data)
        assert found == 0

        pass  # wait for shutdown


# clock = trio.testing.MockClock(rate=1.5,autojump_threshold=0.2)
trio.run(main, sys.argv[1] if len(sys.argv) > 1 else "gpio-mockup-A")  # , clock=clock)



# command line interface

import anyio
import asyncclick as click
from collections.abc import Mapping

from distkv.util import yprint, attrdict, as_service, P
from distkv.data import res_delete, res_get, res_update

import logging

logger = logging.getLogger(__name__)


@click.group(short_help="Manage GPIO controllers.")
async def cli():
    """
    List GPIO controllers, modify device handling …
    """
    pass


@cli.command()
@click.argument("path", nargs=1)
@click.pass_obj
async def dump(obj, path):
    """Emit the current state as a YAML file.
    """
    res = {}
    path = P(path)
    if len(path) > 4:
        raise click.UsageError("Only up to four path elements allowed")

    async for r in obj.client.get_tree(
        obj.cfg.gpio.prefix + path, nchain=obj.meta, max_depth=4 - len(path)
    ):
        # pl = len(path) + len(r.path)
        rr = res
        if r.path:
            for rp in r.path:
                rr = rr.setdefault(rp, {})
        rr["_"] = r if obj.meta else r.value
    yprint(res, stream=obj.stdout)


@cli.command("list")
@click.argument("path", nargs=1)
@click.pass_obj
async def list_(obj, path):
    """List the next stage.
    """
    res = {}
    path = P(path)
    if len(path) > 4:
        raise click.UsageError("Only up to four path elements allowed")

    res = await obj.client._request(
        action="enumerate", path=obj.cfg.gpio.prefix + path, empty=True
    )
    for r in res.result:
        print(r, file=obj.stdout)


@cli.command("attr")
@click.option("-a", "--attr", multiple=True, help="Attribute to list or modify.")
@click.option("-v", "--value", help="New value of the attribute.")
@click.option("-e", "--eval", "eval_", is_flag=True, help="The value shall be evaluated.")
@click.option("-p", "--path", "path_", is_flag=True, help="The value is a path.")
@click.argument("path", nargs=1)
@click.pass_obj
async def attr_(obj, attr, value, path, eval_, path_):
    """Set/get/delete an attribute on a given GPIO element.

    `--eval` without a value deletes the attribute.
    """
    path = P(path)
    if path_ and eval_:
        raise click.UsageError("split and eval don't work together.")
    if value and not attr:
        raise click.UsageError("Values must have locations ('-a ATTR').")
    if path_:
        value = P(value)
    await _attr(obj, attr, value, path, eval_)


@cli.command()
@click.option("-t", "--type", "typ", help="Port type. 'input' or 'output'.")
@click.option("-m", "--mode", help="Port mode. Use '-' to disable.")
@click.option(
    "-a",
    "--attr",
    nargs=2,
    multiple=True,
    help="One attribute to set (NAME VALUE). May be used multiple times.",
)
@click.argument("path", nargs=1)
@click.pass_obj
async def port(obj, path, typ, mode, attr):
    """Set/get/delete port settings. This is a shortcut for the "attr" command.

    \b
    Known attributes for types+modes:
      input:
        read: dest (path)
        count: read + interval (float), count (+-x for up/down/both)
        button: read + t_bounce (float), t_idle (float), skip (+- ignore noise?),
                       t_clear (float), flow (bool)
      output:
        write: src (path), state (path)
        oneshot: write + t_on (float), state (path)
        pulse:   oneshot + t_off (float)
      *:
        low: bool (signals are active-low if true)

    \b
    Paths elements are separated by spaces.
    "low" is the state of the wire when the input is False.
    Floats may be paths, in which case they're read from there when starting.
    """
    path = P(path)
    if len(path) != 3:
        raise click.UsageError("Path must be 3 elements: host gpioname linenr")
    res = await obj.client.get(obj.cfg.gpio.prefix + path, nchain=obj.meta or 1)
    val = res.get("value", attrdict())

    if type is None:
        raise click.UsageError("Port type is mandatory.")
    if mode is None:
        raise click.UsageError("Port mode is mandatory.")
    attr = (("type", typ), ("mode", mode)) + attr
    for k, v in attr:
        if k == "count":
            if v == "+":
                v = True
            elif v == "-":
                v = False
            elif v in "xX*":
                v = None
            else:
                raise click.UsageError("'%s' wants one of + - X" % (k,))
        elif k in ("low", "skip", "flow"):
            if v == "+":
                v = True
            elif v == "-":
                v = False
            else:
                raise click.UsageError("'%s' wants one of + -" % (k,))
        elif k in {"src", "dest"} or (v is not None and " " in v):
            v = v.split()
        else:
            try:
                v = int(v)
            except ValueError:
                try:
                    v = float(v)
                except ValueError:
                    pass
        val[k] = v

    await _attr(obj, (), val, path, False, res)


async def _attr(obj, attr, value, path, eval_, res=None):
    # Sub-attr setter. (Or whole-attr-setter if 'attr' is empty.)
    # Special: if eval_ is True, a value of '-' deletes. A mapping replaces instead of updating.
    if res is None:
        res = await obj.client.get(obj.cfg.gpio.prefix + path, nchain=obj.meta or 2)
    try:
        val = res.value
    except AttributeError:
        res.chain = None
    if eval_:
        if value is None:
            value = res_delete(res, attr)
        else:
            value = eval(value)  # pylint: disable=eval-used
            if isinstance(value, Mapping):
                # replace
                value = res_delete(res, attr)
                value = value._update(attr, value=value)
            else:
                value = res_update(res, attr, value=value)
    else:
        if value is None:
            if not attr and obj.meta:
                val = res
            else:
                val = res_get(res, attr)
            yprint(val, stream=obj.stdout)
            return
        value = res_update(res, attr, value=value)
    res = await obj.client.set(
        obj.cfg.gpio.prefix + path, value=value, nchain=obj.meta, chain=res.chain
    )
    if obj.meta:
        yprint(res, stream=obj.stdout)


@cli.command()
@click.argument("name", nargs=1)
@click.argument("controller", nargs=-1)
@click.pass_obj
async def monitor(obj, name, controller):
    """Stand-alone task to monitor a single contoller.

    The first argument must be the local host name.
    """
    from distkv_ext.gpio.task import task
    from distkv_ext.gpio.model import GPIOroot

    server = await GPIOroot.as_handler(obj.client)
    await server.wait_loaded()
    sub = server[name]
    if controller:
        sub = (sub[x] for x in controller)
    async with as_service(obj) as s:
        async with anyio.create_task_group() as tg:
            e = []
            for chip in sub:
                evt = anyio.create_event()
                await tg.spawn(task, chip, evt)
                e.append(evt)
            for evt in e:
                await evt.wait()
            await s.set()



"""
DistKV client data model for GPIO
"""
import anyio
try:
    from anyio import ClosedResourceError
except ImportError:
    from anyio.exceptions import ClosedResourceError

from distkv.obj import ClientEntry, ClientRoot
from distkv.util import PathLongener, NotGiven
from distkv.errors import ErrorRoot
from distkv.exceptions import ServerError
import asyncgpio as gpio

import logging

logger = logging.getLogger(__name__)


def _DIR(d):
    if d is None:
        return gpio.REQUEST_EVENT_BOTH_EDGES
    if d:
        return gpio.REQUEST_EVENT_RISING_EDGE
    else:
        return gpio.REQUEST_EVENT_FALLING_EDGE


class _GPIObase(ClientEntry):
    """
    Forward ``_update_chip`` calls to child entries.
    """

    _chip = None

    @property
    def chip(self):
        if self._chip is None:
            self._chip = self.parent.chip
        return self._chip

    async def set_value(self, val):  # pylint: disable=arguments-differ
        await super().set_value(val)
        if self.chip is not None:
            await self._update_chip()

    async def update_chip(self):
        await self.parent.update_chip()

    async def _update_chip(self):
        if not self.val_d(True, "present"):
            return
        await self.setup()
        for k in self:
            await k._update_chip()

    async def setup(self):
        pass


class _GPIOnode(_GPIObase):
    """
    Base class for a single input or output.
    """

    _poll = None

    @property
    def card(self):
        return self._path[-2]

    @property
    def port(self):
        return self._path[-1]

    @property
    def task_group(self):
        return self.parent.task_group

    async def setup(self):
        await super().setup()
        if self.chip is None:
            self._poll = None
            return

        if self._poll is not None:
            await self._poll.cancel()
            self._poll = None

        pass


class GPIOline(_GPIOnode):
    """Describes one GPIO line.
    """

    _work = None
    _task_scope = None
    _task_done = None

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.logger = logging.getLogger(".".join(("gpio", self._path[-2], str(self._path[-1]))))

    async def set_value(self, value):  # pylint: disable=arguments-differ
        await super().set_value(value)
        if value is NotGiven:
            await self._kill_task()

    async def _kill_task(self):
        if self._task_scope is not None:
            await self._task_scope.cancel()
            await self._task_done.wait()
            self._task_scope = None
            self._task_done = None

    async def setup(self):
        await super().setup()
        if self.chip is None:
            return
        if self._work:
            await self._work.aclose()

        try:
            typ = self.find_cfg("type")
        except KeyError:
            typ = "NOT_SET"
        if typ == "input":
            await self._setup_input()
        elif typ == "output":
            await self._setup_output()
        else:
            await self.root.err.record_error(
                "gpio",
                self.subpath,
                comment="Line type not set",
                data={"path": self.subpath, "typ": typ}
            )

    async def _task(self, p, *a, **k):
        await self._kill_task()
        try:
            async with anyio.open_cancel_scope() as sc:
                self._task_scope = sc
                self._task_done = anyio.create_event()
                await p(*a, **k)
        finally:
            self._task_scope = None
            await self._task_done.set()

    # Input #

    async def _poll_task(self, evt, dest):
        negate = self.find_cfg("low")
        change = self.find_cfg("change", default=None)
        skip = self.find_cfg("skip")
        bounce = self.find_cfg("t_bounce")

        async with anyio.open_cancel_scope() as sc:
            self._poll = sc
            wire = self.chip.line(self._path[-1])
            with wire.monitor(gpio.REQUEST_EVENT_BOTH_EDGES) as mon:
                old_value = mon.value

                async def set_value(value):
                    if negate:
                        value = not value
                    if change is None or value == change:
                        await self.client.set(dest, value=value)

                await evt.set()
                mon_iter = mon.__aiter__()
                while True:
                    e = await mon_iter.__anext__()

                    if not skip:
                        # assume changed. E.g. old==0, so new==1, val:=1 if negate==0
                        await set_value(not old_value)

                    try:
                        async with anyio.fail_after(bounce):
                            while True:
                                e = await mon_iter.__anext__()
                    except TimeoutError:
                        pass

                    if old_value == e.value:
                        # Some bouncery ended up where it started from.
                        if not skip:
                            # inverse of the above
                            await set_value(old_value)
                    else:
                        if skip:  # otherwise we already sent that
                            await set_value(e.value)
                        old_value = e.value

                    await self.root.err.record_working("gpio", self.subpath)

    async def _button_task(self, evt, dest):
        negate = self.find_cfg("low")
        skip = self.find_cfg("skip")
        bounce = self.find_cfg("t_bounce")
        idle = self.find_cfg("t_idle")
        idle_h = self.find_cfg("t_idle_on", default=idle)
        idle_clear = self.find_cfg("t_clear")
        count = self.find_cfg("count")
        flow = self.find_cfg("flow")

        self.logger.debug("bounce %s idle %s count %s", bounce, idle, count)
        async with anyio.open_cancel_scope() as sc:
            self._poll = sc

            wire = self.chip.line(self._path[-1])
            with wire.monitor(gpio.REQUEST_EVENT_BOTH_EDGES) as mon:
                self.logger.debug("Init %s", mon.value)
                await evt.set()
                ival = None
                mon_iter = mon.__aiter__()

                def td(a, b):
                    a = a.timestamp
                    b = b.timestamp
                    return a[0] - b[0] + (a[1] - b[1]) / 1000000000

                def ts(a):
                    a = a.timestamp
                    return "%03d.%05d" % (a[0] % 1000, a[1] / 10000)

                def inv(x):
                    nonlocal negate

                    if x is None:
                        return x
                    if negate:
                        return not x
                    else:
                        return bool(x)

                # inverting the conditions (and the results, below) is less work
                # than inverting the input values
                count = inv(count)
                ival = mon.value

                async def record(e1, ival):
                    # We record a single sequence of possibly-dirty signals.
                    # e0/e1: first+last change of a sequence with change intervals
                    #        shorter than `bounce`
                    # e2: first change after that sequence has settled, or None when timed out.
                    #
                    res = []
                    e1.value = None
                    e0 = e1
                    self.logger.debug("Start %s %s", e1.value, ts(e1))

                    e2 = None
                    flow_bounce = flow
                    while True:
                        # start with debouncing
                        try:
                            async with anyio.fail_after(bounce):
                                e2 = await mon_iter.__anext__()
                        except TimeoutError:
                            pass
                        else:
                            e1 = e2
                            if flow_bounce and td(e1, e0) > bounce:
                                flow_bounce = False
                                await self.client.set(
                                    dest,
                                    value={
                                        "start": inv(ival),
                                        "seq": res + [0],
                                        "end": inv(e1.value),
                                        "t": bounce,
                                        "flow": True,
                                    }
                                )
                            continue

                        if e2 is None:
                            # didn't bounce at all? good.
                            e2 = e1

                        # If first time, assume we started from Idle.
                        if e0.value is None:
                            e0.value = not ival

                        if e1.value != e0.value:
                            # e0>e1 ends up where it started from, thus we had a dirty signal.
                            if skip or td(e1, e0) < bounce:
                                # ignore it. "Didn't happen."
                                self.logger.debug("Skip: %s %s", ts(e0), ts(e1))
                                e1 = e0
                            else:
                                # treat e1 as legitimate
                                if count is not bool(e1.value):
                                    self.logger.debug("Add+: %s %s", ts(e0), ts(e1))
                                    res.append(int(td(e1, e0) / bounce))
                                    if flow:
                                        await self.client.set(
                                            dest,
                                            value={
                                                "start": not ival,
                                                "seq": res,
                                                "end": inv(e1.value),
                                                "t": bounce,
                                                "flow": True,
                                            }
                                        )
                                else:
                                    self.logger.debug("NoAdd+: %s %s", ts(e0), ts(e1))
                                e0 = e1

                        # Now wait until timeout, or next signal
                        try:
                            async with anyio.fail_after(
                                (idle_h if inv(e1.value) else idle) - bounce
                            ):
                                e2 = await mon_iter.__anext__()
                        except TimeoutError:
                            if count is not bool(e1.value):
                                # We have an infinite signal
                                res.append(0)
                                self.logger.debug("AddZero")
                            return ival, inv(ival), res, inv(e1.value)

                        if count is not (not bool(e1.value)):
                            # First change towards vs. first change away
                            self.logger.debug("Add: %s %s", ts(e0), ts(e2))
                            res.append(int(td(e2, e0) / bounce))
                        else:
                            self.logger.debug("NoAdd: %s %s", ts(e0), ts(e2))

                        # the new signal's value is not yet debounced, so assume for the moment
                        # that it's a change.
                        e2.value = not e1.value
                        e0 = e1 = e2
                        if flow:
                            await self.client.set(
                                dest,
                                value={
                                    "start": not ival,
                                    "seq": res,
                                    "end": inv(e2.value),
                                    "t": bounce,
                                    "flow": True,
                                }
                            )
                        flow_bounce = flow

                clear = True
                while True:
                    if clear and idle_clear:
                        try:
                            async with anyio.fail_after(idle_clear):
                                e = await mon.__anext__()
                        except TimeoutError:
                            await self.client.set(dest, value=False)
                            clear = False
                            continue
                    else:
                        e = await mon.__anext__()
                    ival, start_val, res, end_val = await record(e, ival)
                    if not res:
                        continue

                    await self.client.set(
                        dest, value={"start": start_val, "seq": res, "end": end_val, "t": bounce}
                    )
                    await self.root.err.record_working("gpio", self.subpath)
                    clear = True

    async def _count_task(self, evt, dest):
        intv = self.find_cfg("interval")
        direc = self.find_cfg("count")
        bounce = self.find_cfg("t_bounce")

        async with anyio.open_cancel_scope() as sc:
            self._poll = sc

            async def get_value():
                val = await self.client.get(dest, nchain=2)
                ch = {}
                if "value" in val:
                    ch["chain"] = val.chain

                val = val.get("value", 0)
                if not isinstance(val, (int, float)):
                    val = 0
                return val, ch

            async def set_value():
                nonlocal d, ch, val, t
                self.logger.debug("SEND %d", d)
                t = await anyio.current_time()
                try:
                    res = await self.client.set(dest, value=val + d, nchain=2, **ch)
                    ch["chain"] = res.chain
                except ServerError as exc:
                    # Somebody probably changed my value. Retry once.
                    self.logger.debug("NOSEND %d", d)
                    try:

                        val, ch = await get_value()
                        res = await self.client.set(dest, value=val + d, nchain=2, **ch)
                    except ServerError:
                        await self.root.err.record_error(
                            "gpio",
                            self.subpath,
                            comment="Server error",
                            data={"path": self.subpath, "value": val, **ch},
                            exc=exc
                        )
                    else:
                        self.logger.debug("DIDSEND %d", d)

                ch["chain"] = res.chain
                await self.root.err.record_working("gpio", self.subpath)
                val += d
                d = 0

            val, ch = await get_value()  # pylint: disable=unused-variable ## nonsense

            wire = self.chip.line(self._path[-1])
            # We debounce manually, thus we need both edges
            with wire.monitor(_DIR(None)) as mon:
                await evt.set()
                i = mon.__aiter__()
                t = await anyio.current_time()
                d = None
                value = mon.value
                debounce = True

                # The idea here is that when d is None, nothing is happening, thus we don't time
                # out. Otherwise we send the value at first change, then every `intv` seconds.
                while True:
                    if d is None:
                        await i.__anext__()
                        # reading the value happens later

                        debounce = True
                        t = await anyio.current_time()
                        tm = intv
                    else:
                        tm = max(0, t + intv - await anyio.current_time())
                    try:
                        async with anyio.fail_after(bounce if debounce else tm):
                            await i.__anext__()
                            # reading the value happens later
                    except TimeoutError:
                        # Nothing happened before the timeout.
                        if debounce:
                            debounce = False
                            if mon.value != value:
                                value = mon.value
                                if direc is not (not value):
                                    if d is None:
                                        d = 1
                                        # first change
                                        await set_value()
                                    else:
                                        d += 1
                                        self.logger.debug("UP D %d", d)
                        elif d == 0:
                            d = None
                            # Nothing continues to happen. Skip sending.
                        else:
                            await set_value()
                    else:
                        debounce = True

    async def _setup_input(self):
        try:
            mode = self.find_cfg("mode")
            dest = self.find_cfg("dest")
        except KeyError as exc:
            await self.root.err.record_error(
                "gpio",
                self.subpath,
                comment="mode or dest not set",
                data={"path": self.subpath},
                exc=exc
            )
            return

        evt = anyio.create_event()
        if mode == "read":
            await self.task_group.spawn(self._task, self._poll_task, evt, dest)
        elif mode == "count":
            # These two are in the global config and thus can't raise KeyError
            await self.task_group.spawn(self._task, self._count_task, evt, dest)
        elif mode == "button":
            # These two are in the global config and thus can't raise KeyError
            await self.task_group.spawn(self._task, self._button_task, evt, dest)
        else:
            await self.root.err.record_error(
                "gpio",
                self.subpath,
                comment="mode unknown",
                data={"path": self.subpath, "mode": mode}
            )
            return
        await evt.wait()

    # Output #

    async def with_output(self, evt, src, proc, *args):
        """
        Task that monitors one entry and writes its value to the GPIO controller.

        Also the value is mirrored to ``cur`` if that's set.

        `proc` is called with the GPIO line, the current value from DistKV,
        and the remaining arguments as given.
        """
        async with anyio.open_cancel_scope() as sc:
            self._poll = sc
            with self.chip.line(self._path[-1]).open(direction=gpio.DIRECTION_OUTPUT) as line:
                async with self.client.watch(src, min_depth=0, max_depth=0, fetch=True) as wp:
                    pl = PathLongener()
                    old_val = None
                    await evt.set()
                    async for msg in wp:
                        pl(msg)
                        try:
                            val = msg.value
                        except AttributeError:
                            if msg.get("state", "") != "uptodate":
                                await self.root.err.record_error(
                                    "gpio",
                                    self.subpath,
                                    comment="Missing value in msg",
                                    data={"path": self.subpath, "msg": msg}
                                )
                            continue

                        if val in (False, True, 0, 1):
                            val = bool(val)
                            if False and old_val is val:  # it's a command
                                continue
                            old_val = val
                            try:
                                await proc(line, val, *args)
                            except StopAsyncIteration:
                                await self.root.err.record_error(
                                    "gpio",
                                    self.subpath,
                                    data={"value": val},
                                    comment="Stopped due to bad timer value"
                                )
                                return
                            except Exception as exc:
                                await self.root.err.record_error(
                                    "gpio", self.subpath, data={"value": val}, exc=exc
                                )
                            else:
                                await self.root.err.record_working("gpio", self.subpath)
                        else:
                            await self.root.err.record_error(
                                "gpio", self.subpath, comment="Bad value: %r" % (val,)
                            )

    async def _set_value(self, line, value, state, negate):
        """
        Task that monitors one entry and writes its value to the GPIO controller.

        Also the value is mirrored to ``cur`` if that's set.
        """
        change = self.find_cfg("change", default=None)

        if negate:
            value = not value
        if line is not None:
            self.logger.debug("Setting %s to %s", line, value)
            line.value = value != negate
        if isinstance(value, bool) and change == (not value):
            return
        if state is not None:
            await self.client.set(state, value=value)

    async def _oneshot_value(
        self, line, val, state, negate, t_on
    ):  # pylint: disable=unused-argument
        """
        Task that monitors one entry. Its value is written to the
        controller but if it's = ``direc`` it's reverted autonomously after
        ``intv`` seconds. The current state is written to ``cur``, if
        present.

        ``t_on`` may be a number or a path, if the latter
        it's read from DistKV. ``state`` must be a path.

        """

        async def work_oneshot(evt):
            nonlocal t_on
            if isinstance(t_on, (list, tuple)):
                t_on = (await self.client.get(t_on)).value_or(None)
            async with anyio.open_cancel_scope() as sc:
                try:
                    w, self._work = self._work, sc
                    if w is not None:
                        await w.cancel()
                    try:
                        await self._set_value(line, True, state, negate)
                        await evt.set()
                        await anyio.sleep(t_on)

                    finally:
                        if self._work is sc:
                            async with anyio.fail_after(2, shield=True):
                                await self._set_value(line, False, state, negate)
                finally:
                    await evt.set()  # safety
                    if self._work is sc:
                        self._work = None

        if val:
            evt = anyio.create_event()
            await self.task_group.spawn(work_oneshot, evt)
            await evt.wait()
        else:
            w, self._work = self._work, None
            if w is not None:
                await w.cancel()
                await self._set_value(line, False, state, negate)

    async def _pulse_value(
        self, line, val, state, negate, t_on, t_off
    ):  # pylint: disable=unused-argument
        """
        Pulse the value.

        The state records the cycle ratio.
        """

        async def work_pulse(evt):
            nonlocal t_on
            nonlocal t_off
            if isinstance(t_on, (list, tuple)):
                t_on = (await self.client.get(t_on)).value_or(None)
            if isinstance(t_off, (list, tuple)):
                t_off = (await self.client.get(t_off)).value_or(None)
            if t_on is None or t_off is None:
                raise StopAsyncIteration

            async with anyio.open_cancel_scope() as sc:
                try:
                    w, self._work = self._work, sc
                    if w is not None:
                        await w.cancel()
                    if state is not None:
                        await self.client.set(state, value=t_on / (t_on + t_off))
                    while True:
                        line.value = not negate
                        await evt.set()
                        await anyio.sleep(t_on)
                        line.value = negate
                        await anyio.sleep(t_off)
                finally:
                    await evt.set()
                    if self._work is sc:
                        self._work = None
                        async with anyio.fail_after(2, shield=True):
                            try:
                                line.value = negate
                            except ClosedResourceError:
                                pass
                            else:
                                if state is not None:
                                    await self.client.set(state, value=False)

        if val:
            evt = anyio.create_event()
            await self.task_group.spawn(work_pulse, evt)
            await evt.wait()
        else:
            w, self._work = self._work, None
            if w is not None:
                await w.cancel()
            await self._set_value(line, False, state, negate)

    async def _setup_output(self):

        try:
            mode = self.find_cfg("mode")
            src = self.find_cfg("src")
        except KeyError:
            logger.info("mode or src not set in %s", self.subpath)
            return

        # Rest state. The input value in DistKV is always active=high.
        negate = self.find_cfg("low")
        t_on = self.find_cfg("t_on", default=None)
        t_off = self.find_cfg("t_off", default=None)
        state = self.find_cfg("state", default=None)

        evt = anyio.create_event()
        if mode == "write":
            await self.task_group.spawn(
                self._task, self.with_output, evt, src, self._set_value, state, negate
            )
        elif mode == "oneshot":
            if t_on is None:
                await self.root.err.record_error(
                    "gpio", self.subpath, comment="t_on not set", data={"path": self.subpath}
                )
                return
            await self.task_group.spawn(
                self._task, self.with_output, evt, src, self._oneshot_value, state, negate, t_on
            )
        elif mode == "pulse":
            if t_on is None:
                await self.root.err.record_error(
                    "gpio", self.subpath, comment="t_on not set", data={"path": self.subpath}
                )
                return
            if t_off is None:
                await self.root.err.record_error(
                    "gpio", self.subpath, comment="t_off not set", data={"path": self.subpath}
                )
                return
            await self.task_group.spawn(
                self._task,
                self.with_output,
                evt,
                src,
                self._pulse_value,
                state,
                negate,
                t_on,
                t_off,
            )
        else:
            await self.root.err.record_error(
                "gpio",
                self.subpath,
                comment="mode unknown",
                data={"path": self.subpath, "mode": mode}
            )
            return
        await evt.wait()


class GPIOchip(_GPIObase):
    _chip = None

    @property
    def chip(self):
        return self._chip

    @property
    def name(self):
        return self._path[-1]

    @classmethod
    def child_type(cls, name):
        if not isinstance(name, int):
            return None
        return GPIOline

    async def update_chip(self):
        await self._update_chip()

    async def set_chip(self, chip):
        self._chip = chip
        await self._update_chip()

    async def set_value(self, val):
        await super().set_value(val)
        await self.update_chip()


class GPIOhost(ClientEntry):
    def child_type(self, name):
        return GPIOchip


class GPIOroot(ClientRoot):
    CFG = "gpio"
    err = None
    _chip = None

    async def run_starting(self):
        self._chip = None
        if self.err is None:
            self.err = await ErrorRoot.as_handler(self.client)
        await super().run_starting()

    def child_type(self, name):
        return GPIOhost



from setuptools import setup

LONG_DESC = open("README.rst").read()

setup(
    name="distkv_gpio",
    use_scm_version={"version_scheme": "guess-next-dev", "local_scheme": "dirty-tag"},
    description="A distributed no-master key-value store",
    url="https://github.com/smurfix/distgpio",
    long_description=LONG_DESC,
    author="Matthias Urlichs",
    author_email="matthias@urlichs.de",
    license="MIT -or- Apache License 2.0",
    packages=["distkv_ext.gpio"],
    setup_requires=["setuptools_scm", "pytest-runner", "trustme >= 0.5"],
    install_requires=["distkv >= 0.40", "asyncgpio >= 0.4"],
    tests_require=["trustme >= 0.5", "pytest", "flake8 >= 3.7"],
    keywords=["async", "key-values", "distributed"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Framework :: AsyncIO",
        "Framework :: Trio",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Home Automation",
        "Topic :: System :: Distributed Computing",
    ],
    zip_safe=True,
)
